import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import javax.swing.ImageIcon;


public class GameMap implements Serializable
{
    public static final int ROWS = 10;

    public static final int COLS = 14;

    public static final int X_OFFSET = -18;

    public static final int Y_OFFSET = 0;

    public static final Color[] slimeColors = new Color[] {
            new Color(0, 0, 255),
            new Color(255, 0, 0),
            new Color(0, 0, 255)
    };
    
    private Block[][] mainMap = new Block[ROWS][COLS];
    
    public GameMap(int level)
    {
        try
        {
            BufferedReader levelFile = new BufferedReader(new FileReader(
                    new File("level" + level + ".txt")));
            for(int i = 0; i < ROWS; i++)
            {
                String line = levelFile.readLine();
                for(int j = 0; j < COLS; j++)
                {
                    char block = line.charAt(j);
                    if(block == '#')
                    {
                        setBlockAt(i, j, new Wall());
                    }
                    else if(block == '.')
                    {
                        setBlockAt(i, j, new Block());
                    }
                    else if('0' <= block && block <= '9')
                    {
                        Color c = slimeColors[block - '0'];
                        setBlockAt(i, j, new Slime(block - '0', false, c));
                    }
                    else if('a' <= block && block <= 'j')
                    {
                        Color c = slimeColors[block - 'a'];
                        setBlockAt(i, j, new Slime(block - 'a', true, c));
                    }
                    else if(block == 'p')
                    {
                        setBlockAt(i, j, new Portal());
                    }
                    else
                    {
                        System.err.println(block + " not supported");
                    }
                }
            }
            levelFile.close();
        }
        catch(IOException ioException)
        {
            ioException.printStackTrace();
        }
    }

    public GameMap(GameMap originalMap)
    {
        for(int i = 0; i < ROWS; i++)
        {
            for(int j = 0; j < COLS; j++)
            {
                setBlockAt(i, j, originalMap.getBlockAt(i, j).getCopy());
            }
        }
    }

    public void setBlockAt(int x, int y, Block b)
    {
        mainMap[x][y] = b;
    }
    
    public Block getBlockAt(int x, int y)
    {
        return mainMap[x][y];
    }
    
    private boolean checkSameType(int x, int y, int type)
    {
        return 0 <= x && x < ROWS && 0 <= y && y < COLS && 
                (getBlockAt(x, y) instanceof Slime) &&
                ((Slime)getBlockAt(x, y)).getType() == type;
    }
    
    public void paint(Graphics g, Image face, Image accessory)
    {
        for(int i = 0; i < ROWS; i++)
        {
            for(int j = 0; j < COLS; j++)
            {
                int x = j * Block.WIDTH + X_OFFSET, y = i * Block.HEIGHT + Y_OFFSET;
                if(getBlockAt(i, j) instanceof Slime)
                {
                    BufferedImage slimeImage = new BufferedImage(90, 90, BufferedImage.TYPE_INT_ARGB);
                    Graphics2D g2 = slimeImage.createGraphics();
                    String color = "";
                    Slime s = (Slime)getBlockAt(i, j);
                    if(s.getColor() == slimeColors[0])
                    {
                        color = "Blue";
                    }
                    else if(s.getColor() == slimeColors[1])
                    {
                        color = "Red";
                    }
                    else if(s.getColor() == slimeColors[2])
                    {
                        color = "Green";
                    }
                    g2.drawImage(new ImageIcon(color + "/Middle.png").getImage(), 0, 0, null);
                    g2.drawImage(new ImageIcon(color + "/Top.png").getImage(), 0, 0, null);
                    g2.drawImage(new ImageIcon(color + "/Bottom.png").getImage(), 0, 0, null);
                    g2.drawImage(new ImageIcon(color + "/Left.png").getImage(), 0, 0, null);
                    g2.drawImage(new ImageIcon(color + "/Right.png").getImage(), 0, 0, null);
                    
                    if(checkSameType(i - 1, j, s.getType()) || 
                            checkSameType(i, j - 1, s.getType()))
                    {
                        g2.drawImage(new ImageIcon(color + "/TopLeftSquare.png").getImage(), 0, 0, null);
                    }
                    else
                    {
                        g2.drawImage(new ImageIcon(color + "/TopLeft.png").getImage(), 0, 0, null);
                    }

                    if(checkSameType(i - 1, j, s.getType()) || 
                            checkSameType(i, j + 1, s.getType()))
                    {
                        g2.drawImage(new ImageIcon(color + "/TopRightSquare.png").getImage(), 0, 0, null);
                    }
                    else
                    {
                        g2.drawImage(new ImageIcon(color + "/TopRight.png").getImage(), 0, 0, null);
                    }

                    if(checkSameType(i + 1, j, s.getType()) || 
                            checkSameType(i, j + 1, s.getType()))
                    {
                        g2.drawImage(new ImageIcon(color + "/BottomRightSquare.png").getImage(), 0, 0, null);
                    }
                    else
                    {
                        g2.drawImage(new ImageIcon(color + "/BottomRight.png").getImage(), 0, 0, null);
                    }
                    
                    if(checkSameType(i + 1, j, s.getType()) || 
                            checkSameType(i, j - 1, s.getType()))
                    {
                        g2.drawImage(new ImageIcon(color + "/BottomLeftSquare.png").getImage(), 0, 0, null);
                    }
                    else
                    {
                        g2.drawImage(new ImageIcon(color + "/BottomLeft.png").getImage(), 0, 0, null);
                    }
                    
                    g2.drawImage(face, 0, 0, null);
                    g2.drawImage(accessory, 0, 0, null);
                    g2.dispose();
                    g.drawImage(slimeImage.getScaledInstance(Block.WIDTH, Block.HEIGHT, 0), x, y, null);
                }

                getBlockAt(i, j).paint(g, x, y);
            }
        }
    }
    
    public int[] findBlockAt(int x, int y)
    {
        for(int i = 0; i < ROWS; i++)
        {
            for(int j = 0; j < COLS; j++)
            {
                int curX = j * Block.WIDTH + X_OFFSET;
                int curY = i * Block.HEIGHT + Y_OFFSET;
                if(curX <= x && x < curX + Block.WIDTH &&
                   curY <= y && y < curY + Block.HEIGHT)
                {
                    return new int[] {i, j};
                }
            }
        }
        return new int[] {-1, -1};
    }
    
    public ArrayList<int[]> findSlimeGroup(int x, int y)
    {
        boolean[][] visited = new boolean[ROWS][COLS];
        ArrayList<int[]> group = new ArrayList<int[]>();
        group.add(new int[] {x, y});
        visited[x][y] = true;
        int type = ((Slime)getBlockAt(x, y)).getType();
        for(int i = 0; i < group.size(); i++)
        {
            int curX = group.get(i)[0];
            int curY = group.get(i)[1];
            for(int deltaX = -1; deltaX <= 1; deltaX++)
            {
                for(int deltaY = -1; deltaY <= 1; deltaY++)
                {
                    if(Math.abs(deltaX) + Math.abs(deltaY) == 1)
                    {
                        int newX = curX + deltaX;
                        int newY = curY + deltaY;
                        if(0 <= newX && newX < ROWS && 
                           0 <= newY && newY < COLS && 
                           getBlockAt(newX, newY) instanceof Slime &&
                           ((Slime)getBlockAt(newX, newY)).getType() == type && 
                           !visited[newX][newY])
                        {
                            group.add(new int[] {newX, newY});
                            visited[newX][newY] = true;
                        }
                    }
                }
            }
        }
        return group;
    }

    public boolean linearSearch(ArrayList<int[]> arr, int[] key)
    {
        for(int i = 0; i < arr.size(); i++)
        {
            if(Arrays.equals(arr.get(i), key))
            {
                return true;
            }
        }
        return false;
    }
    
    private Block getActualBlockAt(int x, int y, int movementX, int movementY)
    {
        if(getBlockAt(x + movementX, y + movementY) instanceof Portal)
        {
            for(int i = 0; i < ROWS; i++)
            {
                for(int j = 0; j < COLS; j++)
                {
                    if((getBlockAt(i, j) instanceof Portal) && 
                        (i != x + movementX || j != y + movementY))
                    {
                        return getBlockAt(i + movementX, j + movementY);
                    }
                }
            }
        }
        return getBlockAt(x + movementX, y + movementY);
    }
    
    private void setActualBlockAt(int x, int y, int movementX, int movementY, Block block)
    {
        if(getBlockAt(x + movementX, y + movementY) instanceof Portal)
        {
            for(int i = 0; i < ROWS; i++)
            {
                for(int j = 0; j < COLS; j++)
                {
                    if((getBlockAt(i, j) instanceof Portal) && 
                        (i != x + movementX || j != y + movementY))
                    {
                        setBlockAt(i + movementX, j + movementY, block);
                        return;
                    }
                }
            }
        }
        else
        {
            setBlockAt(x + movementX, y + movementY, block);
        }
    }
    
    public boolean moveOnce(int x, int y, int movementX, int movementY)
    {
        ArrayList<int[]> slimeGroup = findSlimeGroup(x, y);
        boolean validMove = true;
        for(int i = 0; i < slimeGroup.size() && validMove; i++)
        {
            int curX = slimeGroup.get(i)[0];
            int curY = slimeGroup.get(i)[1];
            if((!linearSearch(slimeGroup, new int[] {curX + movementX, curY + movementY}) &&
                  !getActualBlockAt(curX, curY, movementX, movementY).isEmpty()) || 
                ((Slime)getBlockAt(curX, curY)).getAnchored())
            {
                validMove = false;
            }
        }
        if(validMove)
        {
            class CompareCoordinates implements Comparator<int[]>
            {
                private int movement;
                
                public CompareCoordinates(int movement)
                {
                    this.movement = movement;
                }
                
                public int compare(int[] a, int[] b)
                {
                    int retVal;

                    if(a[0] == b[0] && a[1] == b[1])
                    {
                        retVal = 0;
                    }
                    else if(a[0] < b[0])
                    {
                        retVal = 1;
                    }
                    else if(a[0] > b[0])
                    {
                        retVal = -1;
                    }
                    else if(a[1] < b[1])
                    {
                        if(this.movement == 0)
                        {
                            retVal = -1;
                        }
                        else
                        {
                            retVal = this.movement;
                        }
                    }
                    else
                    {
                        if(this.movement == 0)
                        {
                            retVal = 1;
                        }
                        else
                        {
                            retVal = -this.movement;
                        }
                    }

                    return retVal;
                }
            }
            Collections.sort(slimeGroup, new CompareCoordinates(movementY));
            for(int i = 0; i < slimeGroup.size(); i++)
            {
                int curX = slimeGroup.get(i)[0];
                int curY = slimeGroup.get(i)[1];
                Block temp = getBlockAt(curX, curY);
                setBlockAt(curX, curY, getActualBlockAt(curX, curY, movementX, movementY));
                setActualBlockAt(curX, curY, movementX, movementY, temp);
            }
        }
        return validMove;
    }
    
    public void move(int x, int y, int movement)
    {
        if(getBlockAt(x, y) instanceof Slime)
        {
            boolean doGravity = moveOnce(x, y, 0, movement);
            y += movement;
            while(doGravity)
            {
                doGravity = false;
                for(int i = ROWS - 1; i >= 0; i--)
                {
                    for(int j = COLS - 1; j >= 0; j--)
                    {
                        if(getBlockAt(i, j) instanceof Slime)
                        {
                            if(moveOnce(i, j, 1, 0))
                            {
                                doGravity = true;
                            }
                        }
                    }
                }
                x++;
            }
        }
    }
    
    public boolean clickLeft(int x, int y)
    {
        int[] location = findBlockAt(x, y);
        if(location[0] != -1 && location[1] != -1)
        {
            move(location[0], location[1], -1);
            return true;
        }
        else
        {
            return false;
        }
    }

    public boolean clickRight(int x, int y)
    {
        int[] location = findBlockAt(x, y);
        if(location[0] != -1 && location[1] != -1)
        {
            move(location[0], location[1], 1);
            return true;
        }
        else
        {
            return false;
        }
    }
    
    @Override
    public int hashCode()
    {
        int code = 0;
        for(int i = 0; i < ROWS; i++)
        {
            for(int j = 0; j < COLS; j++)
            {
                code = code * 1337 + getBlockAt(i, j).hashCode();
            }
        }
        return code;
    }
    
    @Override
    public boolean equals(Object other)
    {
        if(getClass() != other.getClass())
        {
            return false;
        }
        GameMap otherMap = (GameMap)other;
        for(int i = 0; i < ROWS; i++)
        {
            for(int j = 0; j < COLS; j++)
            {
                if(!getBlockAt(i, j).equals(otherMap.getBlockAt(i, j)))
                {
                    return false;
                }
            }
        }
        return true;
    }
}
