import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Queue;
import java.util.TreeSet;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;

public class GamePanel extends JPanel implements MouseListener, KeyListener
{
    public static final int[] bestMoves = new int[] {12, 14, 10, 25, 9, 12, 12, 24, 24, 16};
    private int level;
    private PlayerProfile profile;
    private GameMap map;
    private ArrayList<GameMap> history;
    private int currentHistory = 0;
    private int exitX = 600, exitY = 40, exitR = 30;
    private int saveX = 500, saveY = 40, saveR = 30;
    private int undoX = 40, undoY = 400, undoR = 30;
    private int redoX = 600, redoY = 400, redoR = 30;
    private ImageIcon exitImage = new ImageIcon("Exit.png");
    private ImageIcon saveImage = new ImageIcon("Save.png");
    private ImageIcon undoImage = new ImageIcon("UndoButton.png");
    private ImageIcon redoImage = new ImageIcon("RedoButton.png");
    
    public GamePanel(int level, PlayerProfile profile)
    {
        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

        this.level = level;
        this.profile = profile;
        
        if(profile.getMap() != null)
        {
            map = new GameMap(profile.getMap());
        }
        else
        {
            map = new GameMap(level);
        }
        
        history = new ArrayList<GameMap>();
        history.add(new GameMap(map));

        addKeyListener(this);
        addMouseListener(this);
    }
    
    public void addNotify()
    {
        super.addNotify();
        requestFocus();
    }
    
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);

        map.paint(g, 
                new ImageIcon("Faces/" + profile.getFace() + ".png").getImage(), 
                new ImageIcon("Accessories/" + profile.getAccessory() + ".png").getImage());
        
        g.drawImage(exitImage.getImage(), exitX - exitR, exitY - exitR, null);
        g.drawImage(saveImage.getImage(), saveX - saveR, saveY - saveR, null);
        g.drawImage(undoImage.getImage(), undoX - undoR, undoY - undoR, null);
        g.drawImage(redoImage.getImage(), redoX - redoR, redoY - redoR, null);
    }
    
    public void mousePressed(MouseEvent e)
    {
        int dxE = e.getX() - exitX;
        int dyE = e.getY() - exitY;
        int dxS = e.getX() - saveX;
        int dyS = e.getY() - saveY;
        int dxU = e.getX() - undoX;
        int dyU = e.getY() - undoY;
        int dxR = e.getX() - redoX;
        int dyR = e.getY() - redoY;
        if(dxE * dxE + dyE * dyE <= exitR * exitR)
        {
            profile.setActiveLevel(0);
            profile.setMap(null);
            SlimeCollection.switchFrames(this, new LevelSelectPanel(profile, ""));
            return;
        }
        else if(dxS * dxS + dyS * dyS <= saveR * saveR)
        {
            profile.setActiveLevel(level);
            profile.setMap(map);
            PlayerProfile.save(profile);
            System.out.println("Game Saved");
        }
        else if(dxU * dxU + dyU * dyU <= undoR * undoR)
        {
            if(currentHistory > 0)
            {
                currentHistory--;
                map = new GameMap(history.get(currentHistory));
            }
        }
        else if(dxR * dxR + dyR * dyR <= redoR * redoR)
        {
            if(currentHistory + 1 < history.size())
            {
                currentHistory++;
                map = new GameMap(history.get(currentHistory));
            }
        }
        else if(SwingUtilities.isLeftMouseButton(e))
        {
            if(map.clickLeft(e.getX(), e.getY()))
            {
                history.subList(currentHistory + 1, history.size()).clear();
                history.add(new GameMap(map));
                currentHistory++;
            }
        }
        else if(SwingUtilities.isRightMouseButton(e))
        {
            if(map.clickRight(e.getX(), e.getY()))
            {
                history.subList(currentHistory + 1, history.size()).clear();
                history.add(new GameMap(map));
                currentHistory++;
            }
        }

        repaint();
        
        if(checkWinCondition(map))
        {
            doWinAction();
        }
    }
    
    public void mouseReleased(MouseEvent e)
    {
        //
    }
    
    public void mouseEntered(MouseEvent e)
    {
        //
    }
    
    public void mouseExited(MouseEvent e)
    {
        //
    }
    
    public void mouseClicked(MouseEvent e)
    {
        //
    }
    
    public void keyPressed(KeyEvent e)
    {
        //
    }

    private boolean checkWinCondition(GameMap checkMap)
    {
        TreeSet<Integer> typesSeen = new TreeSet<Integer>();
        
        class Position implements Comparable<Position>
        {
            public int x, y;
            
            public Position(int x, int y)
            {
                this.x = x;
                this.y = y;
            }
            
            public int compareTo(Position other)
            {
                if(this.x != other.x)
                {
                    return this.x - other.x;
                }
                return this.y - other.y;
            }
        }

        TreeSet<Position> positionsSeen = new TreeSet<Position>();

        boolean win = true;
        
        for(int i = 0; i < GameMap.ROWS && win; i++)
        {
            for(int j = 0; j < GameMap.COLS; j++)
            {
                if(!positionsSeen.contains(new Position(i, j)) &&
                   checkMap.getBlockAt(i, j) instanceof Slime)
                {
                    int type = ((Slime)(checkMap.getBlockAt(i, j))).getType();
                    if(typesSeen.contains(type))
                    {
                        win = false;
                        break;
                    }
                    typesSeen.add(type);
                    for(int[] pos: checkMap.findSlimeGroup(i, j))
                    {
                        positionsSeen.add(new Position(pos[0], pos[1]));
                    }
                }
            }
        }

        return win;
    }
    
    public int[] findImmediateMove(GameMap currentMap, GameMap nextMap)
    {
        for(int i = 0; i < GameMap.ROWS; i++)
        {
            for(int j = 0; j < GameMap.COLS; j++)
            {
                for(int k = -1; k <= 1; k += 2)
                {
                    GameMap nMap = new GameMap(currentMap);
                    nMap.move(i, j, k);
                    if(nMap.equals(nextMap))
                    {
                        return new int[] {i, j, k};
                    }
                }
            }
        }
        return null;
    }
    
    public int[] findNextMove(GameMap currentMap)
    {
        if(checkWinCondition(currentMap))
        {
            return null;
        }

        HashMap<GameMap, GameMap> seen = new HashMap<GameMap, GameMap>();
        seen.put(new GameMap(currentMap), null);
        
        Queue<GameMap> queue = new LinkedList<GameMap>();
        queue.add(new GameMap(currentMap));
        
        while(!queue.isEmpty())
        {
            Thread.yield();
            GameMap cMap = queue.poll();
            
            for(int i = 0; i < GameMap.ROWS; i++)
            {
                for(int j = 0; j < GameMap.COLS; j++)
                {
                    if(cMap.getBlockAt(i, j) instanceof Slime)
                    {
                        for(int k = -1; k <= 1; k += 2)
                        {
                            GameMap nextMap = new GameMap(cMap);
                            nextMap.move(i, j, k);
                            if(!seen.containsKey(nextMap))
                            {
                                seen.put(nextMap, cMap);
                                if(checkWinCondition(nextMap))
                                {
                                    int pathlen = 1;
                                    GameMap nMap = nextMap;
                                    GameMap pMap = seen.get(nMap);
                                    while(!pMap.equals(currentMap))
                                    {
                                        pathlen++;
                                        nMap = pMap;
                                        pMap = seen.get(nMap);
                                    }
                                    System.out.println(pathlen);
                                    return findImmediateMove(pMap, nMap);
                                }
                                queue.add(nextMap);
                            }
                        }
                    }
                }
            }
        }
        return null;
    }
    
    public void doWinAction()
    {
        String userName = JOptionPane.showInputDialog("Please enter your name:");
        
        Score sc = new Score();
        sc.loadFile();
        
        int s = Math.min(bestMoves[level - 1], history.size() - 1 - bestMoves[level - 1]);

        double p = 1.0 - (double)s / bestMoves[level - 1];
        
        int actualScore = 50 + (int)(50 * p * p);
        
        profile.setLevel(Math.max(profile.getLevel(), level + 1));
        profile.setActiveLevel(0);
        profile.setMap(null);
        profile.setMoney(profile.getMoney() + actualScore * level);
        
        sc.renew(level, userName, actualScore);
        
        PlayerProfile.save(profile);

        SlimeCollection.switchFrames(this, new LevelSelectPanel(profile, "You beat stage " + level + " with a score of " + actualScore + "!"));
    }
    
    public void keyReleased(KeyEvent e)
    {
        if(e.getKeyCode() == KeyEvent.VK_X)
        {
            System.out.println("Searching for next move...");
            int[] nextMove = findNextMove(map);
            
            if(nextMove == null)
            {
                System.out.println("No move found");
                JOptionPane.showMessageDialog(null, "The game is impossible to solve. Please undo a move.", "Cheat Mode", JOptionPane.INFORMATION_MESSAGE);
            }
            else
            {
                System.out.println("Next move found");
                map.move(nextMove[0], nextMove[1], nextMove[2]);
                history.subList(currentHistory + 1, history.size()).clear();
                history.add(new GameMap(map));
                currentHistory++;
            }
            
            repaint();
            
            if(checkWinCondition(map))
            {
                doWinAction();
            }
        }
    }
    
    public void keyTyped(KeyEvent e)
    {
        //
    }
}
